#include "global.h"
#include "flags.h"
#include "functions.h"
#include "regs.h"
#include "opcode.h"

typedef struct u_node {
  INDEX instr;
  struct u_node *next;
} *uptr;

typedef struct du_node {
  INDEX defn;
  uptr uselist;
  struct du_node *next;
} *duptr;

typedef struct rd_node {
  INDEX bbl;
  uptr in, out;
  struct rd_node *next;
} *rd_ptr;
 
#define Inst(x)     (x->instr)
#define Defn(x)    (x->defn)
#define UseList(x) (x->uselist)
#define Next(x)    (x->next)
#define In(x)      (x->in)
#define Out(x)     (x->out)
#define BBlk(x)    (x->bbl)

#define AinsUsesReg(ii, r)    (AinsUse(ii) & r)
#define AinsDefinesReg(ii, r) (AinsDef(ii) & r)
#define NewNode(x) (malloc(sizeof(*x)))


#define NO    0
#define MAYBE 1
#define YES   2

#define ABBL_FOREACH_LOOP_IMM_SUBLOOP(loopheader, subloopheader) \
	 ABBL_FOREACH_LOOP_MEMBER(loopheader, subloopheader)      \
	    if(! ABBL_IS_LOOP_HEADER(subloopheader)) continue; else

#define ABBL_FOREACH_LOOP_IMM_BBL(loopheader, bbl)   \
	 ABBL_FOREACH_LOOP_MEMBER(loopheader, bbl)    \
	    if(ABBL_IS_LOOP_HEADER(bbl)) continue; else

#define IS_FUNCTION_CALL(instr) ( (AINS_CODE(instr) == I_jsr) || \
				  (AINS_CODE(instr) == I_bsr) )


/*=====================================================================
  =====================================================================*/

static int no_of_moves, no_of_moves_hot, no_missed, no_missed_hot;
static int no_of_preheaders;
static REGSET availregs;

static BOOL LoopIsWellBehaved(INDEX loophdr);
static REGSET RegsAvailableInLoop(INDEX loophdr);
static BOOL AInsUnaffectedByFuncall(INDEX instr, INDEX ii);
static BOOL LoopDefinesRegs(INDEX loophdr, REGSET r);
static REG FindUnusedReg(INDEX loophdr, REG r);
static NUMBER lub(NUMBER x, NUMBER y) ;
/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ "MakePreHeader"

INDEX MakePreHeader(INDEX loopheader)
{
   INDEX preheader;
   INDEX pred_edge;
   INDEX moved_pred_edge;
   long prehdr_cnt = 0;

   CASSERT( (! ABBL_IS_FIRST(loopheader)), ("A loopheader was also "
					    "the first BBL of a function") );

   no_of_preheaders++;

   preheader = AbblNew();
   ABBL_TYPE(preheader) = BT_NORM;
   
   ABBL_FUN(preheader)               = ABBL_FUN(loopheader);

   /* Put the preheader in the doubly linked list of all BBLs in the
      function.  It has to go just before the loop header since it will
      just fall through. (This is an invariant of the linked list) */

   if (ABBL_IS_FIRST(loopheader))
      {
	 AFUN_BBL_FIRST(ABBL_FUN(loopheader)) = preheader;
      }
   else
      {
	 ABBL_NEXT(ABBL_PREV(loopheader))  = preheader;
      }
   ABBL_PREV(preheader)              = ABBL_PREV(loopheader);
   ABBL_PREV(loopheader)             = preheader;
   ABBL_NEXT(preheader)              = loopheader;

   AbblMigrateRelocInfo(loopheader, preheader);

   /*
    * The following is NO LONGER true:
    * Move all predecessor edges of the loop header wich are not back
    * edges (i.e. not edges from within the loop) to point to the
    * preheader instead.
    */
   FOREACH_BBL_PRED(loopheader,pred_edge) {
     if (! IsBackEdge(pred_edge) ) {
       prehdr_cnt += AEDG_COUNT(pred_edge);
       moved_pred_edge=AedgNewFancy(AEDG_HEAD(pred_edge),preheader,AEDG_TYPE(pred_edge));
       AEDG_COUNT(moved_pred_edge) = AEDG_COUNT(pred_edge);
       AedgKillFancy(pred_edge);
       AedgLinkFancy(moved_pred_edge);
     }
   }

   /* Make a new edge from the preheader to the loopheader */ 

   AedgLinkFancy( AedgNewFancy(preheader, loopheader, ET_NORM) );

   /* fix dominator fields */ 

   ABBL_PRE_DOM(preheader)           = ABBL_PRE_DOM(loopheader);
   ABBL_PRE_DOM(loopheader)          = preheader;

   /* The preheader belongs to the same loop as loopheader. */

   ABBL_LOOP_HEADER(preheader)       = ABBL_LOOP_HEADER(loopheader);
   ABBL_LOOP_NEXT_MEMBER(preheader)  = ABBL_LOOP_FIRST_MEMBER(ABBL_LOOP_HEADER(preheader));
   ABBL_LOOP_FIRST_MEMBER(ABBL_LOOP_HEADER(preheader)) = preheader;

   ABBL_LOOP_FIRST_MEMBER(preheader) = INVALID;
   /*
   ABBL_DEF(preheader)               = REGSET_THE_EMPTY_SET;
   ABBL_USE(preheader)               = REGSET_THE_EMPTY_SET;
   */
   
   /*
   ABBL_IN(preheader)                = ABBL_IN(loopheader);
   */
   ABBL_OUT(preheader)               = AbblRegsLiveIn(loopheader);
   ABBL_COUNT(preheader)             = prehdr_cnt;
   return preheader;
}

/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ "PrintRegSet"

extern void PrintRegSet(REGSET regset)
{
   BOOL firsttime = TRUE;
   REG reg;

   printf("{");
   REGSET_FOREACH_REG(regset, reg)
   {
      if (firsttime)
	 {
	    printf( "%d", reg );
	    firsttime = FALSE;
	 }
      else
	 {
	    printf( ", %d", reg );
	 }
   }
   printf("}");
}
      
/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ "IsLoopExit"

BOOL IsLoopExit(INDEX bbl, INDEX loopheader,  INDEX subloopheader)
{
   INDEX succ_edge;
   INDEX to_loopheader;
   
   CASSERT( (ABBL_LOOP_HEADER(bbl) == subloopheader),
	    ("Incorrect arguments") );

   FOREACH_BBL_SUCC(bbl, succ_edge)
   {
      to_loopheader = ABBL_LOOP_HEADER((AEDG_TO(succ_edge)));
      
      if ( to_loopheader == subloopheader )
	 {
	    /* The edge is internal to the subloop */
	    
	    continue;
	 }
      else if ( Dominates(loopheader, to_loopheader) )
	 {
	    /* The edge is internal to the mainloop */
	    
	    continue;
	 }
      else if ( to_loopheader == bbl )
	 {
	    /* bbl is a loop header and the edge is an
	       edge into the loop of bbl */
	    
	    continue;
	 }
      else
	 {
	    /* otherwise to_loopheader must be 
	       outside the main loop.  (it can't
	       be in a subloop since bbl is not
	       a loop header. */
	    
	    return TRUE;
	 }
   }
   return FALSE;
}

/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ "DominatesAllLoopExitsSL"

BOOL DominatesAllLoopExitsSL(INDEX bbl, INDEX subloopheader, INDEX loopheader)
{
   /* Return TRUE if all nodes in the subloop denoted by subloopheader
      which are exit nodes of the loop denoted by loopheader are
      dominated by bbl.  subloopheader is assumed to be a subloop (not
      neccesarilly an immediate subloop) of loopheader.  subloopheader
      and loopheader may be the same. */

   INDEX imm_bbl;
   INDEX subsubloopheader;

   /* First, if bbl does not dominate every immediate bbl of the subloop
      which is an exit of the main loop, it does not doiminate all
      exits: */

   ABBL_FOREACH_LOOP_IMM_BBL(subloopheader, imm_bbl)
   {
      if ( IsLoopExit(imm_bbl, loopheader, subloopheader) &&
	   ! Dominates( bbl, imm_bbl) )
	 return FALSE;	 
   }

   /* Then, if bbl does not dominate every bbl in sub loops of
      subloopheader which is an exit of the main loop, it does not
      doiminate all exits: */

   ABBL_FOREACH_LOOP_IMM_SUBLOOP(subloopheader, subsubloopheader)
   {
       if ( ! DominatesAllLoopExitsSL(bbl, subsubloopheader, loopheader) )
	 return FALSE;	 
   }

   return ( ! IsLoopExit(subloopheader,
			 loopheader,
			 ABBL_LOOP_HEADER(subloopheader)) ||
	    Dominates( bbl, subloopheader) );
   
}

/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ "DominatesAllLoopExits"

BOOL DominatesAllLoopExits(INDEX bbl, INDEX loopheader)
{
   return DominatesAllLoopExitsSL(bbl, loopheader, loopheader);   
}


/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ "FunDestroyedRegs"

REGSET FunDestroyedRegs(INDEX fun)
{
   /* Returns the set of registers whose contents are (or might be)
      updated by function fun. */

   return AFUN_REGS_CHANGED(fun);
}

/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ "FunCalledByBBL"

INDEX FunCalledByBBL(INDEX ibbl)
{
    CASSERT( ABBL_TYPE(ibbl) == BT_CALL,("bbl doesn't call a function.") );
    return ABBL_FUN( AEDG_TO(ABBL_EDG_CALL(ibbl)) );
}

/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ ""

INDEX FunCalledByInstr(INDEX instr)
{
   CASSERT( (IS_FUNCTION_CALL(instr)),
	    ("Expected jsr or bsr instruction.") );

   return FunCalledByBBL( AINS_BBL(instr) );
}

/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ "IsDefInLoopExcept"

BOOL IsDefInLoopExcept(REG    reg,
		       INDEX  loopheader,
		       REGSET subloopdefs,
		       INDEX  exceptioninstr)
{
   /* returns TRUE if register 'reg' is defined anywhere in
      the loop denoted by 'loopheader' except by the instruction
      'exceptioninstr', provided that 'subloopdefs' is the set
      of registers defined in subloops of the loop.  'exceptioninstr'
      is assumed not to occur in a subloop. */

   INDEX instr;
   INDEX bbl;

   if (REGSET_IS_MEMBER(reg, subloopdefs))
      return TRUE;

   if( REGSET_IS_MEMBER(reg, AbblRegsDef(loopheader)) )
      {
	 /* Aha! 'reg' is defined in 'loopheader'.  Now make
	    sure that it isn't just 'exceptioninstr'
	    that defines it. */
	 
	 FOREACH_BBL_INS(loopheader, instr)
	 {
	    if ( REGSET_IS_MEMBER(reg, AinsDef(instr)) &&
		 instr != exceptioninstr )
	       return TRUE;
	 }
      }

   if ( IS_FUNCTION_CALL(ABBL_INS_LAST(loopheader)) &&
	REGSET_IS_MEMBER(reg, FunDestroyedRegs(FunCalledByBBL(loopheader))) )
      {
	 return TRUE;
      }

   ABBL_FOREACH_LOOP_IMM_BBL(loopheader, bbl)
   {
      if( REGSET_IS_MEMBER(reg, AbblRegsDef(bbl)) )
	 {
	    /* Aha! 'reg' is defined in 'bbl'.  Now make
	       sure that it isn't just 'exceptioninstr'
	       that defines it. */
	    
	    FOREACH_BBL_INS(bbl, instr)
	    {
	       if ( REGSET_IS_MEMBER(reg, AinsDef(instr)) &&
		    instr != exceptioninstr )
		  return TRUE;
	    }
	 }

      if ( IS_FUNCTION_CALL(ABBL_INS_LAST(bbl)) &&
	   REGSET_IS_MEMBER(reg, FunDestroyedRegs(FunCalledByBBL(bbl))) )
	 {
	    return TRUE;
	 }
	 
   }
   return FALSE;
}

/*=====================================================================
  =====================================================================*/
/*
 * LoopDefinesRegs() takes a loop header and a set of registers, and returns
 *  TRUE if any of the registers is defined anywhere in that loop.
 */
#undef __PROC__
#define __PROC__ "LoopDefinesRegs"

static BOOL LoopDefinesRegs(INDEX loophdr, REGSET r)
{
    INDEX bbl;

    ABBL_FOREACH_LOOP_MEMBER(loophdr, bbl)
    {
        if( r & AbblRegsDef(bbl) )
        {
            return FALSE;
        }
    }
    return TRUE;
}

/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ "InvariantOperands"

BOOL InvariantOperands(INDEX instr, REGSET invregs)
{
   /* the operands of an instruction are invariant if either

      - The value of the register defined by the
        instruction has been determined (since this
	means that the operands must have constant values), or

      - All registers used in the instruction are either
        invariant in the loop according to 'invregs', or have a
	known constant value when used in the instruction (it might
	still not be invariant since it can have any other value at
	other point in the loop). */

   if( AINS_FLAGS(instr) & FI_EVAL_C ) return TRUE;
   if( !(AINS_FLAGS(instr) & FI_EVAL_A)  ) return FALSE;
   if( !(AINS_FLAGS(instr) & FI_EVAL_B) ) return FALSE;
   return TRUE;
}

/*=====================================================================
  =====================================================================*/
/*
 * LoadIsLoopInvariant() takes a load instruction, a set of invariant
 * registers and a loopheader, and returns a value indicating whether or
 * not the instruction is invariant within that loop.  There are three
 * possible return values:
 *
 *   YES: indicating that the instruction is definitely loop invariant,
 *        i.e., the memory location loaded from is not changed in the loop
 *         and the register loaded to is invariant in the loop;
 *   MAYBE: indicating that the instruction is potentially invariant, i.e.,
 *        the memory location loaded from is invariant in the loop but the
 *        register loaded to may be changed in the loop; and
 *   NO:  indicating that the instruction is not loop invariant.
 */

#undef __PROC__
#define __PROC__ "AInsUnaffectedByFuncall"

static BOOL AInsUnaffectedByFuncall(INDEX instr, INDEX ii)
{
#if 0
    /*
      rm: this code looks very wrong too me, needs more work
      made uneffective for the time being
     */
    
  NUMBER effect;
  INDEX idef;

  if (!IS_FUNCTION_CALL(ii)) return TRUE;

  effect = AFUN_MOD(FunCalledByInstr(ii));
  if (effect == UNKNOWN)
  {
    return FALSE;
  }
  
  if (AINS_REG_B(idef) == REG_SP && effect == STACK_ONLY)
  {
    /* STACK_ONLY means the callee might write to the stack, not necessarily
       just into its own frame. */
    return FALSE;
  }
  else if( (AINS_FLAGS(instr) & FI_EVAL_B) && effect == GLOBAL_ONLY)
  {
      return FALSE;
  }
  return TRUE;
#else
  return FALSE;
#endif
}

#undef __PROC__
#define __PROC__ "LoadIsLoopInvariant"

long LoadIsLoopInvariant(INDEX instr, REGSET invregs, INDEX loopheader)
{
  INDEX bbl, ii;

  CASSERT(AINS_IS_LOAD(instr), ("argument is not a load instruction\n"));

  if (AINS_TYPE(instr) == IT_LLD) {  /* locked load --- don't move */
    return NO;
  }

  if (!REGSET_IS_MEMBER(AINS_REG_B(instr), invregs)) {
    /* the base register is not invariant in the loop */
    return NO;
  }
  FOREACH_BBL_INS(loopheader, ii) {
    if (IS_FUNCTION_CALL(ii)) {
      if (!AInsUnaffectedByFuncall(instr, ii)) return NO;
    }
    if (AINS_IS_STORE(ii)) {
      if (DoMemoryAccessesConflict(instr, ii) != CONFLICT_NO) { 
	/* possible store to the memory location being loaded from */
	return NO;
      }
    }
  }
  ABBL_FOREACH_LOOP_MEMBER(loopheader, bbl) {
    FOREACH_BBL_INS(bbl, ii) {
      if (IS_FUNCTION_CALL(ii)) {
	if (!AInsUnaffectedByFuncall(instr, ii)) return NO;
      }
      if (AINS_IS_STORE(ii)) {
	if (DoMemoryAccessesConflict(instr, ii) != CONFLICT_NO) {
	  return NO;
	}
      }
    }
  }
  /*
   * At this point we know that the base register is invariant in the loop
   * and the location being loaded from is not modified in the loop.  If the
   * register loaded to is invariant in the loop, the load instruction is
   * itself loop invariant; otherwise it's potentially invariant.
   */
  if (REGSET_IS_MEMBER(AINS_REG_C(instr), invregs)) {
    return YES;
  }
  else {
    return MAYBE;
  }
}
  
/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ "IsMovableInstruction"

long IsMovableInstruction(INDEX instr,
			  REGSET invregs,
			  INDEX loopheader,
			  REGSET subloopdefs)
{
   REG reg;
   REGSET operand_regs;

   if(REGSET_IS_SUBSET(AinsDef(instr),REGSET_UNION(B_ZERO,B_FZERO))) {
     return NO;  /* If all registers defined by 'instr' are zero registers,
		    then this instruction is probably a noop and should not be
		    moved. */
   }
   if ((AINS_TYPE(instr) == IT_BRA) || (AINS_TYPE(instr) == IT_JMP)) {
     return NO;  /* If the instruction is a branch or jump it should not be moved! */
   }
   if ( !REGSET_IS_EMPTY(REGSET_INTERSECT(AinsDef(instr),AbblRegsLiveIn(loopheader))) ) {
     return NO;  /* If a register defined by 'instr' is also defined outside
		    the loop and this definition is used somewhere inside the
		    loop, then the instruction is not movable */
   }
   /* A load instruction can be either loop invariant, or potentially loop
      invariant, or not invariant.
   */
   if ( AINS_IS_LOAD(instr) ) {
     return LoadIsLoopInvariant(instr, invregs, loopheader);
   }
   if ( !InvariantOperands(instr, invregs) ) {
     return NO;  /* If not all  registers used by instr are invariant in the
		    loop, then the instruction is not movable */
   }
   /* If any of the registers defined by instr is also defined somewhere
      else in the loop, then instr is not movable. */

   REGSET_FOREACH_REG(AinsDef(instr), reg) {
     if ( IsDefInLoopExcept(reg, loopheader, subloopdefs, instr) ) {
	return NO;
     }
   }
   /*
    * An instruction can't be moved out of the loop if any of its operands is
    * defined in the loop.  This can happen, for example, if the defining 
    * instruction is a load from memory that happens to be invariant, but can't
    * be moved out of the loop because there are no free registers to hold the
    * value being loaded.
    */
   operand_regs = REG2BIT(AINS_REG_A(instr)) | REG2BIT(AINS_REG_B(instr));
   if (LoopDefinesRegs(loopheader, operand_regs))
   {
     return NO;
   }
   return YES;
}

/*=====================================================================
  =====================================================================*/
/*
 * FindUnusedReg() takes a loop header and returns a register that is not
 * simultaneously live with a given register.
 *
 * The code below does a full-blown liveness analysis every time it's called: 
 * this could surely be improved.  Notice, though, that we can't avoid doing 
 * some sort of liveness analysis, since whenever a dead register is grabbed 
 * for some purpose, liveness information may change.  The right thing to do
 * would be for this information to be updated incrementally.
 */

#undef __PROC__
#define __PROC__ "RegsAvailableInLoop"

static REGSET RegsAvailableInLoop(INDEX loophdr)
{
  INDEX bbl;
  REGSET unavail;

  ComLiveness(stdout);

  unavail = AbblRegsLiveIn(loophdr);
  unavail = REGSET_UNION(unavail, AbblRegsDef(loophdr));
  unavail = REGSET_UNION(unavail, AbblRegsUse(loophdr));
  unavail = REGSET_UNION(unavail, ABBL_OUT(loophdr));
  ABBL_FOREACH_LOOP_MEMBER(loophdr, bbl) {
    unavail = REGSET_UNION(unavail, AbblRegsLiveIn(bbl));
    unavail = REGSET_UNION(unavail, AbblRegsDef(bbl));
    unavail = REGSET_UNION(unavail, AbblRegsUse(bbl));
    unavail = REGSET_UNION(unavail, ABBL_OUT(bbl));
  }
  return REGSET_COMPL(unavail);
}
  
#undef __PROC__
#define __PROC__ "FindUnusedReg"

REG FindUnusedReg(INDEX loophdr, REG r)
{
  REG reg_base, reg_top, i;
  REGLIST avail;

  if (REGSET_IS_MEMBER(r, REGS_INT))
  {
    reg_base = REG_V0;   /* integer register: base is reg 0 */
    reg_top = REG_ZERO;
    avail = REGSET_INTERSECT(availregs, REGS_INT);
  }
  else
  {
    reg_base = REG_F0;   /* float register: base is reg 32 */
    reg_top = REG_FZERO;
    avail = REGSET_INTERSECT(availregs, REGS_FLT);
  }
  
  if (REGSET_IS_EMPTY(avail))
  {
    availregs = RegsAvailableInLoop(loophdr);
  }

  for (i = reg_base; i < reg_top; i++)
  {  /* see if register i is available */
    if (REGSET_IS_MEMBER(i, availregs))
    {
        availregs = REGSET_DIFF(availregs, REGSET_SINGLETON(i));
        return i;
    }
  }
  return REG_ZERO;
}

/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ "PreHeaderOf"

#define ABBL_IS_FALLTHRU(bbl) \
       (VALID(ABBL_FALL_THRU(bbl)) && ! VALID(ABBL_BR_TARGET(bbl)))

extern INDEX PreHeaderOf(INDEX loopheader)
{
   /* Returns the preheader of a loop header.

      If the loop header does not have a unique preheader, a new empty
      preheader is inserted. */

   INDEX pred_edge;
   INDEX preheader;

   CASSERT(ABBL_IS_LOOP_HEADER(loopheader), ("Expected a loop header"));

   preheader = INVALID;

   FOREACH_BBL_PRED(loopheader, pred_edge)
   {
      if( ! IsBackEdge(pred_edge) )
      {
	 if( VALID(preheader) )
	 {
	    /* Oops! it seems loopheader has more than one predecessor
               which is not a back edge, i.e. it does not have a single
	       "pre-header".  We better make one. */

	    return MakePreHeader(loopheader);
	 }
	 else
	 {
	    preheader = AEDG_FROM( pred_edge );
	 }
      }
   }

   if ( VALID(preheader) && ABBL_TYPE(preheader)==BT_NORM )
      return preheader;
   else
      return MakePreHeader(loopheader);
}

/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ "MoveInstr"

void MoveInstr(instr, destination_bbl)
{
   INDEX new_instr;

   if( HasOption(PRINT_OPT_VERBOSE,"Loop") )
   {
       printf("\tMoving Instruction with count %ld in fun %s\n",
              ABBL_COUNT(AINS_BBL(instr)),AFUN_NAME(ABBL_FUN(destination_bbl)));
       PrintIns(stdout,instr);
   }
   
   new_instr = AinsDup(instr);
   AINS_BBL(new_instr) = destination_bbl;
   AinsAppend(new_instr);
   AinsKill(instr);
}

/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ "FindBBLInvariantRegs"

REGSET FindBBLInvariantRegs(INDEX bbl,
			    INDEX loopheader,
			    REGSET invregs,
			    REGSET subloopdefs)
{
  INDEX instr, new_instr, olddest, newdest;
  int is_movable;

  /* move instructions from bbl only if the execution count of bbl is not
     less than that of loopheader. */
  if (ABBL_COUNT(bbl) >= ABBL_COUNT(loopheader)) {
    FOREACH_BBL_INS(bbl, instr) {
      is_movable = IsMovableInstruction(instr,invregs,loopheader,subloopdefs);
      switch (is_movable) {
      case MAYBE:
	if (AINS_IS_LOAD(instr)) {
	  olddest = AINS_REG_C(instr);
	  newdest = FindUnusedReg(loopheader, olddest);
	  if (newdest == REG_ZERO) {  /* no unused register was found */
	    no_missed++; 
	    if (ABBL_IS_HOT(bbl)) no_missed_hot++;
	    break;
	  }
	  /*
	   * replace a load instruction "load rA, ADDR" by a pair of instrs
	   * "load rB, ADDR; rA := 0(rB)", where rB is the free register
	   * returned by FindUnusedReg().  The load instruction will now
	   * presumably be found to be invariant and moved out of the loop.
	   * This effectively replaces a load instruction by a move (lda or cpys)
	   * instruction: it's not worthwhile for cheaper instructions.
	   *
	   * To do this we first replace the destination register of the load
	   * instruction by the free register found, and see if this makes the
	   * load instruction loop invariant: if it does, we insert the "lda"
	   * instruction after it and fall through to the YES case, which then
	   * actually moves it out of the loop; otherwise there's no point 
	   * using up a register and inserting the lda instruction, so we
	   * return the load instruction to its original form.
	   */
	  AINS_REG_C(instr) = newdest;
	  if (IsMovableInstruction(instr,invregs,loopheader,subloopdefs) != YES) {
	    AINS_REG_C(instr) = olddest;  /* original destination */
	    break;
	  }
	  new_instr = AinsNew();
	  if (REGSET_IS_MEMBER(newdest, REGS_INT)) {  /* integer reg: lda */
	    AinsMakeLda(new_instr, newdest, olddest, 0);
	  }
	  else {                                  /* float register: cpys */
	    AINS_CODE(new_instr) = I_cpys;
	    AINS_REG_A(new_instr) = newdest;
	    AINS_REG_B(new_instr) = newdest;
	    AINS_REG_C(new_instr) = olddest;
	  }
	  AINS_BBL(new_instr) = AINS_BBL(instr);
	  AinsInsertAfter(new_instr, instr);
	}
	else {
	  break;
	}
      case YES:
#if 0
printf("===== BEFORE =====\n"); PrintBblSuper(stdout, bbl);
#endif
#if 0
	/**** DEBUG ****/
	printf("    (fun: %s, loop: %d, bbl: %d, count: from = %ld, to = %ld, defs: ", 
	       AFUN_NAME(ABBL_FUN(loopheader)),
	       loopheader, bbl, ABBL_COUNT(bbl), ABBL_COUNT(PreHeaderOf(loopheader)));
	PrintRegSet(AinsDef(instr));
	printf(") ");
	printf(" MOVED!:" );
	PrintIns(stdout,instr);
	/**** DEBUG ****/
#endif
	no_of_moves++;
	if (ABBL_IS_HOT(bbl)) no_of_moves_hot++;
	MoveInstr(instr,PreHeaderOf(loopheader));
	invregs = REGSET_UNION(AinsDef(instr),invregs);
#if 0
printf("===== AFTER =====\n"); 
printf("PREHEADER: \n"); PrintBblSuper(stdout, PreHeaderOf(loopheader));
printf("BBL:\n"); PrintBblSuper(stdout, bbl);
#endif
	break;
      case NO: 
	break;
      default:
	fprintf(stderr, "ERROR: Unrecognized return value (%d) from IsMovableInstruction()\n", 
		is_movable);
	break;
      } /* switch */
    }
  }
  return invregs;
}


/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ "FindLoopInvariantRegs"

REGSET FindLoopInvariantRegs(INDEX loopheader,
			     REGSET invregs,
			     REGSET subloopdefs)
{
   /* Given a set 'invregs' of registers known to be invariant
      in the loop denoted by 'loopheader', this function returns
      a set of registers containing
         - all registers in invregs, and
	 - any register defined in a basic block not appearing in
	   a subloop of this loop which can be determined to be
	   defined with an invariant value.
   */

   REGSET oldinvregs;
   INDEX  bbl;
   BOOL inst_moved = FALSE;

   oldinvregs = REGS_ALL; /* This is just to make the while loop
				    condition succeed the first time. */
   while( invregs != oldinvregs )
   {
      oldinvregs = invregs;
      invregs = FindBBLInvariantRegs(loopheader,loopheader,invregs,subloopdefs);
      /*
       * Process the basic blocks in the loop in two passes: in the
       * first pass, look at the "hot" blocks only; in the second
       * pass, look at all the blocks.  This is an attempt to prevent
       * cold blocks from using up registers and thereby blocking 
       * optimizations out of hot blocks.
       */
      ABBL_FOREACH_LOOP_IMM_BBL(loopheader, bbl)
      {
	if ( ABBL_IS_HOT(bbl) ) {
	  if ( DominatesAllLoopExits(bbl,loopheader) ) {
	    invregs = FindBBLInvariantRegs(bbl,loopheader,invregs,subloopdefs);
	    inst_moved = TRUE;
	  }
	}
      }
      ABBL_FOREACH_LOOP_IMM_BBL(loopheader, bbl)
      {
	if ( DominatesAllLoopExits(bbl,loopheader) ) {
	  invregs = FindBBLInvariantRegs(bbl,loopheader,invregs,subloopdefs);
	  inst_moved = TRUE;
	}
      }
   }
   return invregs;
}

/*=====================================================================
  =====================================================================*/
/*
 * A loop is well behaved if it doesn't contain any basic block that
 * calls Hell, or contains a PAL instruction, or has an escaping edge.
 */
#undef __PROC__
#define __PROC__ "LoopIsWellBehaved"

static BOOL LoopIsWellBehaved(INDEX loophdr)
{
  INDEX bbl, e, hd, tl;
  
  if (ABBL_TYPE(loophdr) == BT_PAL)
  {
    return FALSE;
  }
  
  FOREACH_BBL_SUCC(loophdr, e)
  {
      hd = ABBL_FUN(AEDG_HEAD(e));
      tl = ABBL_FUN(AEDG_TAIL(e));
      if( AFUN_IS_PSEUDO(hd) || AFUN_IS_PSEUDO(tl) )
      {
          return FALSE;
      }
      if (hd != tl && AEDG_TYPE(e) != ET_FUN_CALL
	&& AEDG_TYPE(e) != ET_FUN_RETURN) {  /* escaping edge */
      return FALSE;
    }
  }
  ABBL_FOREACH_LOOP_MEMBER(loophdr, bbl) {
    if (ABBL_TYPE(bbl) == BT_PAL) {
      return FALSE;
    }
    FOREACH_BBL_SUCC(bbl, e) {
      hd = ABBL_FUN(AEDG_HEAD(e));
      tl = ABBL_FUN(AEDG_TAIL(e));
      if( AFUN_IS_PSEUDO(hd) || AFUN_IS_PSEUDO(tl) )
      {
	return FALSE;
      }
      if (hd != tl && AEDG_TYPE(e) != ET_FUN_CALL
	  && AEDG_TYPE(e) != ET_FUN_RETURN) {  /* escaping edge */
	return FALSE;
      }
    }
  }  
  return TRUE;
}
  
/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ "MoveLoopInvariantDefs"

REGSET MoveLoopInvariantDefs(INDEX loopheader)
{
   REGSET subloopdefs;
   INDEX  subloopheader;
   REGSET newdefs;
   REGSET loopdefs;
   REGSET invregs;
   BOOL hdr_preds_ok;
   INDEX  bbl, iedg;

   if (!LoopIsWellBehaved(loopheader))
   {
     return REGS_NONE;
   }

   /* First, move invariant definitions out of subloops.  This also yields
      a set 'subloopdefs' of registers defined in any subloop (after any
      invariant definitions have been moved out). */

   subloopdefs = REGS_NONE;

   ABBL_FOREACH_LOOP_IMM_SUBLOOP(loopheader, subloopheader)
   {
      newdefs = MoveLoopInvariantDefs(subloopheader);
      subloopdefs = REGSET_UNION(subloopdefs,newdefs);
   }
   /*
    * Recompute what registers are available in the loop.  This is
    * necessary because some previously available registers may have
    * been grabbed by sub-loops of this loop when they were processed
    * recursively.  (This also serves to initialize availregs before each
    * loop is processed.)
    */
   availregs = REGS_NONE;

   /* Then, find what registers are defined in the loop
      denoted by 'loopheader', including definitions in
      subloops. */

   loopdefs = subloopdefs;
   loopdefs = REGSET_UNION(loopdefs,AbblRegsDef(loopheader));

   if ( IS_FUNCTION_CALL(ABBL_INS_LAST(loopheader)) )
      {
	 /* If the last instruction of the loopheader is a function call,
	    any register modified by this function is defined in the loop. */

	 loopdefs =
	    REGSET_UNION(loopdefs,
			 FunDestroyedRegs(FunCalledByBBL(loopheader)));
      }

    ABBL_FOREACH_LOOP_IMM_BBL(loopheader, bbl)
    {
      loopdefs = REGSET_UNION(loopdefs, AbblRegsDef(bbl));

      if ( IS_FUNCTION_CALL(ABBL_INS_LAST(bbl)) )
	 {
	    /* If the last instruction of the bbl is a function call then
	       any register destroyed/modified by this function is
	       defined in the loop. */

	    loopdefs =
	       REGSET_UNION(loopdefs,
			    FunDestroyedRegs(FunCalledByBBL(bbl)));
	 }
   }
   
    /*
     * determine whether any of the incoming edges into the loop header
     * is a function call edge or a return edge: if it is, don't do any
     * code motion out of the loop.  This restriction is due to constraints
     * on ways in which edges entering a loop can be retargeted: it turns
     * out that call/return edges can't be retargeted without bad effects
     * down the line, so we steer clear of this situation.
     */
    hdr_preds_ok = TRUE;
    FOREACH_BBL_PRED(loopheader, iedg) {
      if (AEDG_TYPE(iedg) == ET_FUN_CALL ||
	  AEDG_TYPE(iedg) == ET_FUN_RETURN) {
	hdr_preds_ok = FALSE;
	break;
      }
    }
    if (hdr_preds_ok) {
      /* Find all registers which are invariant inside the loop.
	 To start with, all registers *not* defined inside the loop are
	 invariant in the loop */
      
      invregs = REGSET_COMPL(loopdefs);
      
      /* The integer and floating point "zero registers" are always invariant.
	 (B_ZERO is the singleton set contatining the integer
	 zero register, and  B_FZERO is the floatingpoint equivalent.) */

      invregs = REGSET_UNION(invregs, REGSET_UNION(B_ZERO,B_FZERO));
      /* FindLoopInvariantRegs actually moves instructions */
      
      invregs = FindLoopInvariantRegs(loopheader, invregs, subloopdefs);
    }
    return loopdefs;
}

/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ "MoveFunLoopCode"

#define FOREACH_FUN_LOOP(f,loop)\
   FOREACH_FUN_BBL(f,loop)       \
     if (! ABBL_IS_LOOP_HEADER(loop)) continue; else

void MoveFunLoopCode(INDEX f)
{
   INDEX loop;

   FOREACH_FUN_TOPLOOP(f,loop)
   {
     MoveLoopInvariantDefs(loop);
   }
}

/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ "OptMoveLoopInvariantCode"

extern BOOL OptMoveLoopInvariantCode(FILE *fp)
{
   INDEX f;
   BOOL change = 0;
   clock_t time = clock();

   no_of_moves = no_of_moves_hot = no_missed = no_missed_hot = 0;
   no_of_preheaders = 0;

   FOREACH_LIVE_FUN(f)
   {
      MoveFunLoopCode(f);
   }

   change = no_of_moves > 0;
   time = (clock() - time) * 1000 / CLOCKS_PER_SEC;

   InfoStatus(fp,"\t[OptLoop%c] Moved Inss: %d#%d  [Missed: %d#%d] {%d ms}\n",
          CHANGE(change), no_of_moves, no_of_moves_hot, no_missed, no_missed_hot,time);

   return change;
}


/*
 * ComputeFunctionSideEffects() does a simple dataflow analysis to
 * estimate the side effects a function can have.  The possible values
 * are:
 *
 *  NO_SIDE_EFFECT : the function doesn't store to memory;
 *  GLOBAL_ONLY    : only global variables are modified;
 *  LOCAL_ONLY     : only the function's own locals are modified;
 *  STACK_ONLY     : only the stack is modified (i.e., not heap/globals).
 *  UNKNOWN        : anything may be modified.
 *
 * The analysis proceeds in two phases.  In the first phase, we determine
 * the "local" effects of each function, by examining the effects of
 * store instructions in the function body, and also whether the stack size
 * is valid and whether the body contains indirect function calls.
 * These are fixed for each function, and so need be analysed only once.
 * After this, we iteratively propagate side effect information from
 * callee to caller using a standard fixpoint computation.
 */

static NUMBER lub(NUMBER x, NUMBER y) 
{
  if (x == y) return x;
  if (x == NO_SIDE_EFFECT) return y;
  if (y == NO_SIDE_EFFECT) return x;
  if (x <= STACK_ONLY && y <= STACK_ONLY) return STACK_ONLY;
  return UNKNOWN;
}

extern void ComputeFunctionSideEffects(FILE *fp)
{
  INDEX f, bbl, ii,
      e;
  BOOL chg;
  NUMBER effect, tmpeffect, stksz, n_tot, n_none, n_local, n_stack, n_global;
  clock_t time = clock();


  /* initialization */
  
  FOREACH_LIVE_FUN(f)
  {
      if( AFUN_IS_PSEUDO(f) )
      {
          AFUN_MOD(f) = UNKNOWN;
      }
      else
      {
          AFUN_MOD(f) = NO_SIDE_EFFECT;
      }
  }


  /*
   * Phase 1: examine stack size, store instructions, jsrs.
   */
  FOREACH_LIVE_FUN(f) {
    if (AFUN_MOD(f) == UNKNOWN) {
      continue;
    }
    
    stksz = -AFUN_STK_MIN(f);
    if ( AFUN_HAS_STK_INVALID(f) )
    {
      AFUN_MOD(f) = UNKNOWN;
      chg = TRUE;
      continue;
    }
    FOREACH_FUN_BBL(f, bbl) {
      /*
       * check each store instruction in the block, and look for jsr instrs.
       */
      FOREACH_BBL_INS(bbl, ii) {
	if (AINS_CODE(ii) == I_jsr) {
	  AFUN_MOD(f) = UNKNOWN;
	  break;
	}
	if (AINS_IS_STORE(ii))
        {
            if( AINS_FLAGS(ii) & FI_EVAL_B)
            {
                tmpeffect = GLOBAL_ONLY;
            }
            else if (AINS_REG_B(ii) == REG_SP)
            {
                if (AINS_DATA(ii) <= stksz)
                {
                    tmpeffect = LOCAL_ONLY;
                }
                else
                {
                    tmpeffect = STACK_ONLY;
                }
            }
            else
            {
                tmpeffect = UNKNOWN;
            }
            effect = lub(AFUN_MOD(f), tmpeffect);
            if ( effect > AFUN_MOD(f) )
            {
                AFUN_MOD(f) = effect;
            }
	}
      }
    }
  }
  /*
   * Phase 2: iteratively propagate information from called functions
   * to their callers.
   */
  chg = TRUE;
  while (chg) {
    chg = FALSE;
    FOREACH_LIVE_FUN(f) {
      if (AFUN_MOD(f) == UNKNOWN) {
	continue;
      }
      FOREACH_FUN_BBL(f, bbl) {
	FOREACH_BBL_SUCC(bbl, e) {
	  if (ABBL_FUN(AEDG_HEAD(e)) != ABBL_FUN(AEDG_TAIL(e)) 
	      && AEDG_TYPE(e) != ET_FUN_RETURN) {  /* function call or escape edge */
	    effect = lub(AFUN_MOD(AEDG_HEAD(e)), AFUN_MOD(AEDG_TAIL(e)));
	    if (effect > AFUN_MOD(f)) {
	      AFUN_MOD(f) = effect;
	      chg = TRUE;
	    }
	  }
	}
      }
    }
  }
  /* Finally, tally up the results. */
  n_tot = n_none = n_local = n_stack = n_global = 0;
  FOREACH_LIVE_FUN(f) {
    n_tot++;
    switch (AFUN_MOD(f)) {
    case NO_SIDE_EFFECT: 
      n_none++;
      break;
    case LOCAL_ONLY:
      n_local++;
      break;
    case STACK_ONLY:
      n_stack++;
      break;
    case GLOBAL_ONLY:
      n_global++;
      break;
    }
  }
  time = (clock() - time) * 1000 / CLOCKS_PER_SEC;

  InfoStatus(fp, "\t[ComputeSideEffect] Total %d, NoSideEffect %d, Local %d, Stack %d, Global %d {%d ms}\n",
	  n_tot, n_none, n_local, n_stack, n_global,time);

}

