/*=======================================================
  Author:   R. Muth
  Version:  $Id: opt.partialred.c,v 1.2 1998/09/21 23:21:25 muth Exp muth $
  Comment:  This file contains the partial redundancy optimization
=======================================================*/

#include "global.h"
#include "functions.h"

#include "flags.h"
#include "opcode.h"
#include "regs.h"

/*=======================================================
=======================================================*/
static INT64 weight;
/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "AbblUnspeculate"

static BOOL AbblUnspeculate(INDEX ibbl, BOOL verbose)
{
    INDEX iins;
    INDEX ft;
    INDEX bt;
    REGLIST use_later;
    REGLIST def_later;
        
    if( ABBL_TYPE(ibbl) != BT_BRANCH ) return 0;
    
    ft = AEDG_TAIL(ABBL_EDG_FALSE(ibbl));
    bt = AEDG_TAIL(ABBL_EDG_TRUE(ibbl));

    if( bt == ibbl ) return 0;

    def_later = B_ZERO|B_FZERO;
    use_later = REGS_NONE;
    
    
    FOREACH_BBL_INS_R(ibbl,iins)
    {
        REGLIST use = AinsUse(iins);
        REGLIST def = AinsDef(iins);

        if( def & use_later )       /* do not move w-r conflict */
            goto label_next_ins;
        
        if( def & def_later )       /* do not move w-w conflict */
            goto label_next_ins;

        if( use & def_later )       /* do not move r-w conflict */
            goto label_next_ins;

        if( AinsHasSideEffect(iins) ) /* do not move ins with side effects */
            goto label_next_ins;
        
        if( !(def & ABBL_OUT(ibbl)) )     /* do not move dead stuff */
            goto label_next_ins;

        if( AINS_TYPE(iins) == IT_ILD || AINS_TYPE(iins) == IT_FLD )
        {
            INDEX jins;
            for( jins =  ABBL_INS_LAST(ibbl); jins != iins; jins = AINS_PREV(jins) )
            {
                if( AINS_TYPE(jins) == IT_IST || AINS_TYPE(jins) == IT_FST )
                {
                    if( DoMemoryAccessesConflict(iins,jins) != CONFLICT_NO )  /* no conflict ? */
                        goto label_next_ins;
                }
            }
        }
        
        
        /*
          move ins into bt branch if not used in ft branch
          note, that the in-sets might change while more and more instructions are
          moved and hence needs recomputation
          */
        
        if( AbblNpreds(bt)  == 1  && !(def & AbblRegsLiveIn(ft)) )
        {
            INDEX jins = AinsDup(iins);

            weight += ABBL_COUNT(ibbl) - ABBL_COUNT(bt);
            
            AINS_BBL(jins) = bt;
                    
            AinsPrepend(jins);
            AinsKill(iins);
            if( verbose )
            {
                fprintf(stdout,"%d (%ld) -> %d (%ld):",ibbl,ABBL_COUNT(ibbl),bt,ABBL_COUNT(bt));
                PrintIns(stdout,jins);
            }
            return 1;
        }
        
        if( AbblNpreds(ft) == 1 && !(def & AbblRegsLiveIn(bt)) )
        {
            INDEX jins = AinsDup(iins);

            weight += ABBL_COUNT(ibbl) - ABBL_COUNT(ft);
                        
            AINS_BBL(jins) = ft;
            
            AinsPrepend(jins);
            AinsKill(iins);
            if( verbose )
            {

                fprintf(stdout,"%d (%ld) -> %d (%ld):",ibbl,ABBL_COUNT(ibbl),ft,ABBL_COUNT(ft));
                PrintIns(stdout,jins);
            }
            return 1;
        }

label_next_ins:
        use_later |= use;
        def_later |= def;
    }

    return 0;
}

/*=======================================================
@OPT
If a register which is defined in a conditional branch bbl
is only used in one of the branches the defining instruction
may be moved into the branch that actually used the result
under some side constaints.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptUndoSpeculations"

extern BOOL OptUndoSpeculations(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Spec");
    INDEX ibbl;
    NUMBER mcount = 0;
    NUMBER mcounth = 0;
    BOOL change;
    clock_t time = clock();

    weight = 0;  /* global variable */

    
    FOREACH_LIVE_BBL(ibbl)
    {
        while( AbblUnspeculate(ibbl,verbose) )
        {
            if( ABBL_IS_HOT(ibbl) )
                mcounth++;

            mcount++;
        }
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = mcount > 0;

    InfoStatus(fp,"\t[OptSpec%c] Moved Inss: %d#%d  wt: %ld {%d ms}\n",CHANGE(change),mcount,mcounth,weight,time);

    return change;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "OptUndoSpeculations"

extern BOOL ObsoleteOptUndoSpeculations(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Spec");
    INDEX ibbl;
    NUMBER mcount = 0;
    NUMBER mcounth = 0;
    INT64  weight = 0;
    BOOL change;
    clock_t time = clock();
        
    FOREACH_LIVE_BBL(ibbl)
    {
        INDEX iins;
        BOOL  encountered_store;
        REGLIST use_later;
        REGLIST def_later;
        INDEX ft;
        INDEX bt;
        NUMBER  ft_n_preds;
        NUMBER  bt_n_preds;

        if( ABBL_TYPE(ibbl) != BT_BRANCH ) continue;
        
        ft = AEDG_TAIL(ABBL_EDG_FALSE(ibbl));
        bt = AEDG_TAIL(ABBL_EDG_TRUE(ibbl));
        if( bt == ibbl ) continue;
        def_later = B_ZERO|B_FZERO;
        use_later = REGS_NONE;
        ft_n_preds = AbblNpreds(ft);
        bt_n_preds = AbblNpreds(bt);

        encountered_store = 0;   /* NEEDS WORK: should use aliasing information */
    
        FOREACH_BBL_INS_R(ibbl,iins)
        {
            REGLIST use = AinsUse(iins);
            REGLIST def = AinsDef(iins);

            if(
               !(def & use_later) &&       /* do not move w-r conflict */
               !(def & def_later) &&       /* do not move w-w conflict */ 
               !(use & def_later) &&       /* do not move r-w conflict */
               ((AINS_TYPE(iins)!=IT_ILD && AINS_TYPE(iins)!=IT_FLD)  || !encountered_store) && 
               !AinsHasSideEffect(iins) && /* do not move ins with side effects */
               def & ABBL_OUT(ibbl)     /* do not move dead stuff */
               )  
            {
                
                /*
                  move ins into bt branch if not used in ft branch
                  note, that the in-sets might change while more and more instructions are
                  moved and hence needs recomputation
                  */

                if( bt_n_preds == 1  && !(def & AbblRegsLiveIn(ft)) )
                {
                    INDEX jins = AinsDup(iins);

                    AINS_BBL(jins) = bt;
                    mcount++;
                    if( ABBL_IS_HOT(ibbl) )
                        mcounth++;
                    weight += ABBL_COUNT(ibbl) - ABBL_COUNT(bt);
                    
                             
                    if( verbose )
                    {
                        fprintf(stdout,"%d (%ld) -> %d (%ld):",ibbl,ABBL_COUNT(ibbl),bt,ABBL_COUNT(bt));
                        PrintIns(fp,jins);
                        PrintBblSuper(fp,ibbl);
                    }
                    
                    AinsPrepend(jins);
                    AinsKill(iins);
                    continue; /* we moved it, so no book-keeping for this one*/
                }
                
                if( ft_n_preds == 1 && !(def & AbblRegsLiveIn(bt)) )
                {
                    INDEX jins = AinsDup(iins);

                    AINS_BBL(jins) = ft;
                    mcount++;
                    if( ABBL_IS_HOT(ibbl) )
                        mcounth++;
                    weight += ABBL_COUNT(ibbl) - ABBL_COUNT(ft);
                    
                    if( verbose )
                    {
                        fprintf(stdout,"%d (%ld) -> %d (%ld):",ibbl,ABBL_COUNT(ibbl),ft,ABBL_COUNT(ft));
                        PrintIns(fp,jins);
                        PrintBblSuper(fp,ibbl);
                    }
                    
                    AinsPrepend(jins);
                    AinsKill(iins);
                    continue;
                }
            }
            
            
            use_later |= use;
            def_later |= def;
            
            if( AINS_TYPE(iins) == IT_IST ||
                AINS_TYPE(iins) == IT_FST ||
                AINS_TYPE(iins) == IT_CST ||
            AINS_TYPE(iins) == IT_UST )
                encountered_store = 1;
        } /* loop ins */
        
    } /* loop bbls */

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = mcount > 0;

    InfoStatus(fp,"\t[OptSpec%c] Moved Inss: %d hot: %d wt: %ld {%d ms}\n",CHANGE(change),mcount,mcounth,weight,time);

    return change;
}
/*=======================================================
=======================================================*/
