/*=======================================================
  Author:   
  Version:  $Id$
  Comment:  
=======================================================*/


#include "global.h"
#include "flags.h"
#include "functions.h"
#include "regs.h"
#include "opcode.h"

#define NUM_SAVED_REGS  8

#define RegSetToSmap(rset)     ((rset & 0xff00) >> 8)
#define SmapToRegSet(smap)     ((smap<<8) & 0xff00)
#define SmapIntersect(s1, s2)  (s1 & s2)
#define SmapUnion(s1, s2)      (s1 | s2)
#define SmapComplement(s)      (~s)
#define SmapIsSubset(s1,s2)    (SmapUnion(s1,s2) == s2)

#define EdgeIsOrdinary(e)      (AEDG_TYPE(e) == ET_TRUE \
                                || AEDG_TYPE(e) == ET_FALSE \
				|| AEDG_TYPE(e) == ET_UNCOND \
				|| AEDG_TYPE(e) == ET_NORM \
				|| AEDG_TYPE(e) == ET_FUN_LINK)

static int  ninstrs = 0;
static long winstrs = 0;


 /* Anticipated regs at entry: used in shrink-wrapping */
#define ABBL_ANTOUT(i)       ABBL_U1(i).ubyte.b1
/* Available regs at exit: used in shrink-wrapping */
#define ABBL_AVIN(i)         ABBL_U1(i).ubyte.b2
#define ABBL_SAVE(i)         ABBL_U1(i).ubyte.b3
#define ABBL_RESTORE(i)      ABBL_U1(i).ubyte.b4
#define ABBL_IS_SAVED(i)     ABBL_U1(i).ubyte.b5
#define ABBL_IS_RESTORED(i)  ABBL_U1(i).ubyte.b6
#define ABBL_APP(i)          ABBL_U1(i).ubyte.b7






/*=======================================================
@FUN
AbblRedefinesRegs(ibbl) computes the set of registers that are
defined in block ibbl, not counting as definitions the restoring
of registers from memory in any return block of a function.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblRedefinesRegs"

extern REGSET AbblRedefinesRegs(INDEX ibbl)
{
  INDEX iins;
  REGSET iregs;
  REG r;

  CASSERT((VALID(ibbl)), ("invalid argument: %d\n", ibbl));

  if (ABBL_TYPE(ibbl) == BT_EXIT) {
    return REGS_NONE;
  }

  if (ABBL_TYPE(ibbl) != BT_RET) {
    return AbblRegsDef(ibbl);
  }
  /*
   * work back to the last load of the register; then see if
   * that register is redefined earlier in the block.
   */
  iregs = REGS_NONE;

  REGSET_FOREACH_REG(SAVE_REGS, r) {
    FOREACH_BBL_INS_R(ibbl, iins) {
      if ( (AINS_CODE(iins) == I_ldq || AINS_CODE(iins) == I_ldt)
	   && AINS_REG_C(iins) == r) {
#if 0
	CWARNING((AINS_REG_B(iins) != REG_SP),
		("restore not from stack: reg %d [fun=%d]\n", r, ABBL_FUN(ibbl)));
#endif
	break;
      }
    }
    for (iins = AINS_PREV(iins); VALID(iins); iins = AINS_PREV(iins)) {
      if (AINS_REG_C(iins) == r) {
	iregs |= REG2BIT(r);
	break;
      }
    }
  }

  return iregs;
}
	  
/*=======================================================
@FUN
AbblRestoresRegs(ibbl) computes the set of registers that are
restored in block ibbl, i.e., loaded into a register that is
not subsequently modified in ibbl.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblRestoresRegs"

extern REGSET AbblRestoresRegs(INDEX ibbl)
{
  INDEX iins;
  REGSET iregs;
  REG r;

  CASSERT((VALID(ibbl)), ("invalid argument: %d\n", ibbl));

  if (ABBL_TYPE(ibbl) == BT_EXIT) {
    return REGS_NONE;
  }
  /*
   * find the last load of the register.
   */
  iregs = REGS_NONE;
  REGSET_FOREACH_REG(SAVE_REGS, r) {
    FOREACH_BBL_INS_R(ibbl, iins) {
      if ( (AINS_CODE(iins) == I_ldq || AINS_CODE(iins) == I_ldt)
	   && AINS_REG_C(iins) == r) {
#if 0
	CWARNING((AINS_REG_B(iins) != REG_SP),
		("restore not from stack: reg %d [fun=%d]\n", r, ABBL_FUN(ibbl)));
#endif
	iregs |= REG2BIT(r);
	break;
      }
    }
  }

  return iregs;
}

/*=======================================================
@FUN
AbblSavesRegs(ibbl) computes the set of registers that are
saved in block ibbl, i.e., stored into memory prior to any
modification.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblSavesRegs"

extern REGSET AbblSavesRegs(INDEX ibbl)
{
  INDEX iins;
  REGSET iregs;
  REG r;

  CASSERT((VALID(ibbl)), ("invalid argument: %d\n", ibbl));

  if (ABBL_TYPE(ibbl) == BT_EXIT) {
    return REGS_NONE;
  }
  /*
   * find the first store of the register.
   */
  iregs = REGS_NONE;
  REGSET_FOREACH_REG(SAVE_REGS, r) {
    FOREACH_BBL_INS(ibbl, iins) {
      if ( (AINS_CODE(iins) == I_stq || AINS_CODE(iins) == I_stt)
	   && AINS_REG_A(iins) == r) {
#if 0
	CWARNING((AINS_REG_B(iins) != REG_SP),
		("save not to stack: reg %d [fun=%d]\n", r, ABBL_FUN(ibbl)));
#endif
	iregs |= REG2BIT(r);
	break;
      }
    }
  }

  return iregs;
}

/*=======================================================
@FUN
AbblUsesRegs(ibbl) computes the set of registers that are
used in block ibbl, not counting as uses the saving of registers
to memory in the initial block of a function.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblUsesRegs"

extern REGSET AbblUsesRegs(INDEX ibbl)
{
  INDEX iins;
  REGSET iregs;
  REG r;

  CASSERT((VALID(ibbl)), ("invalid argument: %d\n", ibbl));

  if (ABBL_TYPE(ibbl) == BT_EXIT) {
    return REGS_NONE;
  }

  iregs = REGS_NONE;
  if (ibbl != AFUN_BBL_FIRST(ABBL_FUN(ibbl))) {
    FOREACH_BBL_INS(ibbl, iins) {
      iregs |= AinsUse(iins);
    }
    return iregs;
  }
  else {
    REGSET_FOREACH_REG(SAVE_REGS, r) {
      FOREACH_BBL_INS(ibbl, iins) {
	if ( (AINS_CODE(iins) == I_stq || AINS_CODE(iins) == I_stt)
	     && AINS_REG_A(iins) == r) {
#if 0
	  CWARNING((AINS_REG_B(iins) != REG_SP),
		  ("save not to stack: reg %d [fun=%d]\n", r, ABBL_FUN(ibbl)));
#endif
	  break;
	}
      }
      for (iins = AINS_NEXT(iins); VALID(iins); iins = AINS_NEXT(iins)) {
	if (AINS_REG_A(iins) == r) {
	  iregs |= REG2BIT(r);
	  break;
	}
      }
    }
  }

  return iregs;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblAntOut"

static UINT8 AbblAntOut(INDEX ibbl, UINT8 sregmap)
{
  UINT8 antout, antin_succ;
  INDEX iedg, jbbl;

  if (ABBL_TYPE(ibbl) == BT_RET || ABBL_TYPE(ibbl) == BT_EXIT) {
    return REGS_NONE;
  }

  if (ABBL_TYPE(ibbl) == BT_CALL         /* assume all regs saved by */
      || ABBL_TYPE(ibbl) == BT_JUMP      /* this function are used */
      || ABBL_TYPE(ibbl) == BT_PAL) {  
    return sregmap & RegSetToSmap(SAVE_REGS);
  }
  
  antout = sregmap;
  FOREACH_BBL_SUCC(ibbl, iedg) {
    if (!EdgeIsOrdinary(iedg)) continue;
    jbbl = AEDG_TAIL(iedg);
    antin_succ = SmapUnion(ABBL_ANTOUT(jbbl), ABBL_APP(jbbl));
    antout = SmapIntersect(antout, antin_succ);
  }
  return antout;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblAvIn"

static UINT8 AbblAvIn(INDEX ibbl, UINT8 sregmap)
{
  UINT8 avin, avout_pred;
  INDEX iedg, jbbl;

  if (ibbl == AFUN_BBL_FIRST(ABBL_FUN(ibbl))) {
    return REGS_NONE;
  }

  /*
   * If ibbl is a "return block", i.e., the target of a function
   * return, then we know that for the corresponding call block
   * we set the "AntIn" attribute to consist of all the registers
   * saved by this function (see above); so here we set AvOut to
   * be the same set of registers.
   */
  FOREACH_BBL_PRED(ibbl, iedg) {
    if (AEDG_TYPE(iedg) == ET_FUN_LINK) {
      return  sregmap & RegSetToSmap(SAVE_REGS);
    }
  }

  avin = sregmap;
  FOREACH_BBL_PRED(ibbl, iedg) {
    if (!EdgeIsOrdinary(iedg)) continue;
    jbbl = AEDG_HEAD(iedg);
    avout_pred = SmapUnion(ABBL_AVIN(jbbl), ABBL_APP(jbbl));
    avin = SmapIntersect(avin, avout_pred);
  }
  return avin;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblApp"

static UINT8 AbblApp(INDEX ibbl)
{
  UINT8 app, currapp;
  INDEX loophdr, jbbl;

  if (ABBL_TYPE(ibbl) == BT_EXIT) {
    return REGS_NONE;
  }

  if (ABBL_IS_IN_ANY_LOOP(ibbl)) {
    loophdr = ABBL_IS_LOOP_HEADER(ibbl) ? ibbl : ABBL_LOOP_HEADER(ibbl);
    app = RegSetToSmap((AbblRedefinesRegs(loophdr) | AbblUsesRegs(loophdr)));
    ABBL_FOREACH_LOOP_MEMBER(loophdr, jbbl) {
      if (ABBL_IS_LOOP_HEADER(jbbl)) {
	currapp = AbblApp(jbbl);
      }
      else {
	currapp = RegSetToSmap((AbblRedefinesRegs(jbbl) | AbblUsesRegs(jbbl)));
      }
      app = SmapUnion(app, currapp);
    }
  }
  else {
    app = RegSetToSmap((AbblRedefinesRegs(ibbl) | AbblUsesRegs(ibbl)));
  }

  return app;
}    

/*=======================================================
@FUN
AbblPropagateLoopSaveRestore propagates save/restore information
from outside a loop into it, by considering edges where one end
is in the loop and the other end is not in it, and intersecting
the save/restore information from such edges appropriately.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblPropagateLoopSaveRestore"

static BOOL AbblPropagateLoopSaveRestore(INDEX loophdr, UINT8 sregmap)
{
  INDEX ibbl, jbbl, iedg;
  UINT8 saveregs, restoreregs, sregs;
  BOOL  chg = FALSE;

  CASSERT(ABBL_IS_LOOP_HEADER(loophdr), ("argument is not loop header: %d\n", loophdr));

  saveregs = sregmap;
  FOREACH_BBL_PRED(loophdr, iedg) {
    if (!EdgeIsOrdinary(iedg)) continue;
    jbbl = AEDG_HEAD(iedg);
    if (ABBL_LOOP_HEADER(jbbl) == loophdr) continue;  /* intra-loop edge */
    saveregs = SmapIntersect(ABBL_IS_SAVED(jbbl), saveregs);
  }
  
  restoreregs = sregmap;
  FOREACH_BBL_SUCC(loophdr, iedg) {
    if (!EdgeIsOrdinary(iedg)) continue;
    jbbl = AEDG_TAIL(iedg);
    if (ABBL_LOOP_HEADER(jbbl) == loophdr) continue;  /* intra-loop edge */
    restoreregs = SmapIntersect(ABBL_IS_RESTORED(jbbl), restoreregs);
  }

  ABBL_FOREACH_LOOP_MEMBER(loophdr, ibbl) {
    FOREACH_BBL_PRED(loophdr, iedg) {
      if (!EdgeIsOrdinary(iedg)) continue;
      jbbl = AEDG_HEAD(iedg);
      if (ABBL_LOOP_HEADER(jbbl) == loophdr) continue;  /* intra-loop edge */
      saveregs = SmapIntersect(ABBL_IS_SAVED(jbbl), saveregs);
    }
    
    FOREACH_BBL_SUCC(loophdr, iedg) {
      if (!EdgeIsOrdinary(iedg)) continue;
      jbbl = AEDG_TAIL(iedg);
      if (ABBL_LOOP_HEADER(jbbl) == loophdr) continue;  /* intra-loop edge */
      restoreregs = SmapIntersect(ABBL_IS_RESTORED(jbbl), restoreregs);
    }
  }
  /*
   * Now propagate these sets to each block in the loop.
   */
  sregs = SmapUnion(saveregs, ABBL_SAVE(loophdr));
  if (!SmapIsSubset(sregs, ABBL_IS_SAVED(loophdr))) {
    ABBL_IS_SAVED(loophdr) = sregs;
    chg = TRUE;
  }
  sregs = SmapUnion(restoreregs, ABBL_RESTORE(ibbl));
  if (!SmapIsSubset(sregs, ABBL_IS_RESTORED(loophdr))) {
    ABBL_IS_RESTORED(loophdr) = sregs;
    chg = TRUE;
  }
  ABBL_FOREACH_LOOP_MEMBER(loophdr, ibbl) {
    sregs = SmapUnion(saveregs, ABBL_SAVE(ibbl));
  if (!SmapIsSubset(sregs, ABBL_IS_SAVED(ibbl))) {
      ABBL_IS_SAVED(ibbl) = sregs;
      chg = TRUE;
    }
    sregs = SmapUnion(restoreregs, ABBL_RESTORE(ibbl));
  if (!SmapIsSubset(sregs, ABBL_IS_RESTORED(ibbl))) {
      ABBL_IS_RESTORED(ibbl) = sregs;
      chg = TRUE;
    }
  }

  return chg;
}

/*=======================================================
@FUN
AfunShrinkWrapIsSafe examines the save/restore information computed
for a function to see if it's safe.
Shrink-wrapping is safe for a function if two conditions hold:
(1) if a register is saved in a block B, it is not saved again in
    any block reachable from B; if a register is restored in a
    block B, it is not restored in a block reachable from B; and

(2) if a block is saved in a block B, it is restored along all
    paths from B; if a register is restored in a block B, it
    has been saved along all paths to B.

These conditions are somewhat more restrictive than those imposed
by Chow in his PLDI-88 paper.  However, they simplify the code
considerably, and -- more important -- preclude the possibility of
performance loss due to multiple saves/restores along some execution
path (cf. Fig. 3 in Chow's paper).
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunShrinkWrapIsSafe"

static BOOL AfunShrinkWrapIsSafe(INDEX ifun, UINT8 sregmap)
{
  BOOL  verbose = HasOption(PRINT_OPT_VERBOSE,"ShrinkWrap");
  INDEX ibbl, jbbl, iedg, loophdr;
  UINT8 sregs, tmp0;
  BOOL chg;

  FOREACH_FUN_BBL(ifun, ibbl) {
    ABBL_IS_SAVED(ibbl) = ABBL_IS_RESTORED(ibbl) = RegSetToSmap(REGS_NONE);
  }
  chg = TRUE;
  while (chg) {
    chg = FALSE;
    FOREACH_FUN_BBL(ifun, ibbl) {
      if (ABBL_TYPE(ibbl) == BT_EXIT) continue;
      if (ibbl != AFUN_BBL_FIRST(ifun)) {
	tmp0 = RegSetToSmap(REGS_NONE);
	FOREACH_BBL_PRED(ibbl, iedg) {
	  if (EdgeIsOrdinary(iedg)) {
	    jbbl = AEDG_HEAD(iedg);
	    sregs = SmapUnion(ABBL_IS_RESTORED(jbbl), ABBL_RESTORE(jbbl));
	    tmp0 = SmapUnion(sregs, tmp0);
	  }
	}
	if (ABBL_IS_RESTORED(ibbl) != tmp0) {
	  ABBL_IS_RESTORED(ibbl) = tmp0;  /* restored on some path upto ibbl */
	  chg = TRUE;
	}
      }
      
      if (ABBL_TYPE(ibbl) != BT_RET && ABBL_TYPE(ibbl) != BT_EXIT) {
	tmp0 = RegSetToSmap(REGS_NONE);
	FOREACH_BBL_SUCC(ibbl, iedg) {
	  if (EdgeIsOrdinary(iedg)) {
	    jbbl = AEDG_TAIL(iedg);
	    sregs = SmapUnion(ABBL_IS_SAVED(jbbl), ABBL_SAVE(jbbl));
	    tmp0 = SmapUnion(sregs, tmp0);
	  }
	}
	if (ABBL_IS_SAVED(ibbl) != tmp0) {
	  ABBL_IS_SAVED(ibbl) = tmp0;  /* saved on some path from ibbl */
	  chg = TRUE;
	}
      }
    }
  }

  FOREACH_FUN_BBL(ifun, ibbl) {
    if (ABBL_TYPE(ibbl) == BT_EXIT) continue;
    if (SmapIntersect(ABBL_SAVE(ibbl), ABBL_IS_SAVED(ibbl))) {
      /*
       * some register that is being saved here may be saved again later.
       */
      if (verbose) {
	printf("\t\t[ShrinkWrap] bailing out (repeated save): fun %d, bbl %d\n", ifun, ibbl);
      }
      return FALSE;
    }
    if (SmapIntersect(ABBL_RESTORE(ibbl), ABBL_IS_RESTORED(ibbl))) {
      /*
       * some register that is being restored here may be restored again later.
       */
      if (verbose) {
	printf("\t\t[ShrinkWrap] bailing out (repeated restore): fun %d, bbl %d\n", ifun, ibbl);
      }
      return FALSE;
    }
  }      
  /*
   * Next, check condition (2) above.
   *
   * ABBL_IS_SAVED(ibbl) now represents the set of registers saved on
   * all paths from the entry node of ifun upto and including ibbl;
   * ABBL_IS_RESTORED(ibbl) represents the set of registers that
   * are restored on all paths from (including) ibbl to all the exit
   * nodes of ifun.
   */
  FOREACH_FUN_BBL(ifun, ibbl) {
    ABBL_IS_SAVED(ibbl) = ABBL_SAVE(ibbl);
    ABBL_IS_RESTORED(ibbl) = ABBL_RESTORE(ibbl);
  }
  chg = TRUE;
  while (chg) {
    chg = FALSE;
    FOREACH_FUN_BBL(ifun, ibbl) {
      /*
       * First propagate information into loops.
       */
      if (ABBL_TYPE(ibbl) == BT_EXIT || !ABBL_IS_IN_ANY_LOOP(ibbl)) continue;

      loophdr = VALID(ABBL_LOOP_HEADER(ibbl)) ? ABBL_LOOP_HEADER(ibbl) : ibbl;
      chg |= AbblPropagateLoopSaveRestore(loophdr, sregmap);
    }
    FOREACH_FUN_BBL(ifun, ibbl) {
      /*
       * Next propagate information into non-loop blocks.
       */
      if (ABBL_TYPE(ibbl) == BT_EXIT || ABBL_IS_IN_ANY_LOOP(ibbl)) continue;

      if (ibbl != AFUN_BBL_FIRST(ifun)) {
	tmp0 = sregmap;
	FOREACH_BBL_PRED(ibbl, iedg) {
	  if (EdgeIsOrdinary(iedg)) {
	    jbbl = AEDG_HEAD(iedg);
	    tmp0 = SmapIntersect(ABBL_IS_SAVED(jbbl), tmp0);
	  }
	}
	tmp0 = SmapUnion(ABBL_SAVE(ibbl), tmp0);
	if (ABBL_IS_SAVED(ibbl) != tmp0) {
	  ABBL_IS_SAVED(ibbl) = tmp0;  /* saved on all paths upto ibbl */
	  chg = TRUE;
	}
      }
	
      if (ABBL_TYPE(ibbl) != BT_RET && ABBL_TYPE(ibbl) != BT_EXIT) {
	tmp0 = sregmap;
	FOREACH_BBL_SUCC(ibbl, iedg) {
	  if (EdgeIsOrdinary(iedg)) {
	    jbbl = AEDG_TAIL(iedg);
	    tmp0 = SmapIntersect(ABBL_IS_RESTORED(jbbl), tmp0);
	  }
	}
	tmp0 = SmapUnion(ABBL_RESTORE(ibbl), tmp0);
	if (ABBL_IS_RESTORED(ibbl) != tmp0) {
	  ABBL_IS_RESTORED(ibbl) = tmp0;  /* restored on all paths from ibbl */
	  chg = TRUE;
	}
      }
    }
  } /* while chg */
  FOREACH_FUN_BBL(ifun, ibbl) {
    if (ABBL_TYPE(ibbl) == BT_EXIT) continue;

    if (ABBL_IS_SAVED(ibbl) != ABBL_IS_RESTORED(ibbl)) {
      if (verbose) {
	printf("\t\t[ShrinkWrap] bailing out (unbalanced save/restore): fun %d, bbl %d\n", ifun, ibbl);
      }
      return FALSE;
    }
  }
  return TRUE;
}
  
/*=======================================================
@FUN
ComShrinkWrapInfo(ifun) computes the placement of save/restore actions
for callee-save registers in the basic blocks of function ifun.  This 
follows the description given in:

F. C. Chow, "Minimizing Register Usage Penalty at Procedure Calls",
Proc. PLDI-88, pp. 85-94.

This paper describes the meaning and computation of the sets
ABBL_APP, ABBL_ANT{IN,OUT}, and ABBL_AV{IN,OUT} associated with
each basic block.  The main thing to note is that the callee
saved integer registers on the DEC Alpha registers 9-15, and to
conserve space we currently consider only these registers (actually,
we consider registers 8-15).  Information about these registers
is kept in bit vectors of size 8.

At some point we need to look into whether it's worthwhile doing
shrink-wrapping for floating point registers.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "ComShrinkWrapInfo"

static UINT8 ComShrinkWrapInfo(INDEX ifun)
{
  INDEX  ibbl, jbbl, iedg;
  UINT8   sregmap, antin, antout, avin, avout, save, restore;
  BOOL   chg;

  ibbl = AFUN_BBL_FIRST(ifun);
  sregmap = RegSetToSmap(AbblSavesRegs(ibbl));

  FOREACH_FUN_BBL(ifun, ibbl) {
    if (ABBL_TYPE(ibbl) == BT_EXIT) continue;
    ABBL_APP(ibbl) = AbblApp(ibbl);
    if (ibbl == AFUN_BBL_FIRST(ifun) || ABBL_TYPE(ibbl) == BT_RET) {
      ABBL_AVIN(ibbl) =  RegSetToSmap(REGS_NONE);
      ABBL_ANTOUT(ibbl) = RegSetToSmap(REGS_NONE);
    }
    else {
      ABBL_AVIN(ibbl) =  sregmap;
      ABBL_ANTOUT(ibbl) = sregmap;    
    }
  }
  
  chg = TRUE;
  while (chg) {
      chg = FALSE;
      FOREACH_FUN_BBL(ifun, ibbl) {
	if (ABBL_TYPE(ibbl) == BT_EXIT) continue;

	avin = AbblAvIn(ibbl, sregmap);
	if (ABBL_AVIN(ibbl) != avin) {
	  ABBL_AVIN(ibbl) = avin;
	  chg = TRUE;
	}

	antout = AbblAntOut(ibbl, sregmap);
	if (ABBL_ANTOUT(ibbl) != antout) {
	  ABBL_ANTOUT(ibbl) = antout;
	  chg = TRUE;
	}
      }
    }
    /*
     * Now compute SAVE and RESTORE sets for the blocks
     */
    FOREACH_FUN_BBL(ifun, ibbl) {
      if (ABBL_TYPE(ibbl) == BT_EXIT) continue;
      antin   = SmapUnion(ABBL_ANTOUT(ibbl), ABBL_APP(ibbl));
      save    = SmapIntersect(antin, SmapComplement(ABBL_AVIN(ibbl)));
      FOREACH_BBL_PRED(ibbl, iedg) {
	if (EdgeIsOrdinary(iedg)) {
	  jbbl = AEDG_HEAD(iedg);
	  antin = SmapUnion(ABBL_ANTOUT(jbbl), ABBL_APP(jbbl));
	  save = SmapIntersect(save, SmapComplement(antin));
	}
      }
      ABBL_SAVE(ibbl) = save;

      avout   = SmapUnion(ABBL_AVIN(ibbl), ABBL_APP(ibbl));
      restore = SmapIntersect(avout, SmapComplement(ABBL_ANTOUT(ibbl)));
      FOREACH_BBL_SUCC(ibbl, iedg) {
	/*
	 * ignore back edges when computing restore sets: otherwise
	 * registers that are used in the loop are not restored.
	 */
	if (IsBackEdge(iedg)) continue;
	if (EdgeIsOrdinary(iedg)) {
	  jbbl = AEDG_TAIL(iedg);
	  avout = SmapUnion(ABBL_AVIN(jbbl), ABBL_APP(jbbl));
	  restore = SmapIntersect(restore, SmapComplement(avout));
	}
      }
      ABBL_RESTORE(ibbl) = restore;
    }

    return sregmap;
}

/*=======================================================
@FUN
To make sure that the insertion of restore instructions works
properly around function calls, it's convenient to ensure that
each call block Bc has a corresponding block Br such that Br is
the only successor of Bc, and Bc is the only predecessor of Br.
The function AfunInsertRetBbls preprocesses the basic blocks
and inserts such blocks Br as necessary to ensure that this is 
the case.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunInsertRetBbls"

static BOOL AfunInsertRetBbls(INDEX ifun)
{
  INDEX ibbl, jbbl, iedg, calltarget, nextbbl;
  int npreds;

  FOREACH_FUN_BBL(ifun, ibbl) {
    if (ABBL_TYPE(ibbl) != BT_CALL) {
      continue;
    }
    calltarget = AEDG_TAIL(ABBL_EDG_CALL(ibbl));

    if (AFUN_IS_PSEUDO(ABBL_FUN(calltarget))) {
      return FALSE;
    }

    nextbbl = AEDG_TAIL(ABBL_EDG_LINK(ibbl));
    npreds = 0;
    FOREACH_BBL_PRED(nextbbl, iedg) {
      if (!AEDG_IS_INTER(iedg)) {
	npreds++;
      }
    }

    if (npreds > 1) {  /* more than one predecessor in ifun */
      /*
       * Create a new return block and splice it in.
       */
      jbbl = AbblNew();
      ABBL_TYPE(jbbl) = BT_NORM;
      ABBL_COUNT(jbbl) = ABBL_COUNT(ibbl);
      ABBL_FUN(jbbl) = ifun;
      ABBL_RESTORE(jbbl) = ABBL_RESTORE(ibbl);
      AbblAppendFancy(jbbl);
      
      AedgLinkFancy(AedgNewFancy(jbbl,nextbbl,ET_NORM));
      AedgKillFancy( ABBL_EDG_CALL(ibbl));
      AedgKillFancy( ABBL_EDG_LINK(ibbl));
      
      AedgLinkFancy(AedgNewFancy(ibbl,jbbl,ET_FUN_LINK));
      AedgLinkFancy( AedgNewFancy(ibbl,calltarget,ET_FUN_CALL));
    }
  }
  return TRUE;
}

/*=======================================================
@FUN
Carry out shrink-wrapping on a function.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunShrinkWrap"

extern void AfunShrinkWrap(INDEX ifun)
{
  BOOL  verbose = HasOption(PRINT_OPT_VERBOSE,"ShrinkWrap");
  INDEX save_instr[NUM_SAVED_REGS];
  INDEX ibbl, jbbl, iins, xins;
  REGSET sregs;
  REG r, rA, rC;
  int i;
  BOOL chg = FALSE;
  long delta = 0;

  CASSERT(VALID(ifun), ("illegal argument: %d\n", ifun));

  ibbl = AFUN_BBL_FIRST(ifun);
  sregs = SmapToRegSet(RegSetToSmap(AbblSavesRegs(ibbl)));

  for (i = 0; i < NUM_SAVED_REGS; i++) {
    save_instr[i] = -1;
  }

  REGSET_FOREACH_REG(sregs, r) {
    /*
     * map this register to a position in the save map.
     */
    if (r < 8 || r > 15) continue;
    i = r-8;
    /*
     * Now find the save instruction for this reg.
     */
    FOREACH_BBL_INS(ibbl, iins) {
      if ( (AINS_CODE(iins) == I_stq || AINS_CODE(iins) == I_stt)
	   && AINS_REG_A(iins) == r) {
	save_instr[i] = iins;
	break;
      }
    }
  }
  /*
   * We don't move any saves/restores for registers that aren't saved
   * into the stack.
   */
  for (i = 0; i < NUM_SAVED_REGS; i++) {
    iins = save_instr[i];
    if (iins > 0) {
      if (AINS_REG_B(iins) != REG_SP) {
	save_instr[i] = -1;
      }
    }
  }

  FOREACH_FUN_BBL(ifun, ibbl) {
    if (ABBL_TYPE(ibbl) == BT_EXIT) continue;
    if (ibbl == AFUN_BBL_FIRST(ifun)) {
      if (SmapToRegSet(ABBL_SAVE(ibbl)) == AbblSavesRegs(ibbl)) {
	continue;  /* no need to do anything */
      }
    }
    else if (ABBL_TYPE(ibbl) == BT_RET) {
      if (SmapToRegSet(ABBL_RESTORE(ibbl)) == AbblRestoresRegs(ibbl)) {
	continue;  /* no need to do anything */
      }
    }
    else {
      /*
       * insert save instructions, if necessary
       */
      if (ABBL_SAVE(ibbl)) {
	CWARNING(VALID(ABBL_LOOP_HEADER(ibbl)),
		 ("Shrink-wrapping moves saves into a loop[bbl=%d]\n", ibbl));
	REGSET_FOREACH_REG(SmapToRegSet(ABBL_SAVE(ibbl)), r) {
	  i = r-8;
	  if (save_instr[i] < 0) continue;

	  xins = AinsDup(save_instr[i]);
	  AINS_BBL(xins) = ibbl;
	  AinsInsertBefore(xins, ABBL_INS_FIRST(ibbl));
	  winstrs -= ABBL_COUNT(ibbl);
	  delta   -= ABBL_COUNT(ibbl);
	  if (verbose) {
	    printf("\t\t[ShrinkWrap: +save $%d: bbl %d, fun %d (ct = %ld)]\n", r, ibbl, ifun, ABBL_COUNT(ibbl));
	  }
	  chg = TRUE;
	}
      }
      if (ABBL_RESTORE(ibbl)) {
	CWARNING(VALID(ABBL_LOOP_HEADER(ibbl)), ("Shrink-wrapping moves restores into a loop[bbl=%d]\n", ibbl));
	/*
	 * If this is a call block, the restore instruction should
	 * go, not at the end of this block, but into the corresponding
	 * return block.  
	 */
	jbbl = ABBL_TYPE(ibbl) == BT_CALL ? AEDG_TAIL(ABBL_EDG_LINK(ibbl)) : ibbl;
	if (ABBL_TYPE(jbbl) == BT_RET &&
	    SmapToRegSet(ABBL_RESTORE(jbbl)) == AbblRestoresRegs(jbbl)) {
	continue;  /* no need to do anything */
	}

	REGSET_FOREACH_REG(SmapToRegSet(ABBL_RESTORE(jbbl)), r) {
	  i = r-8;
	  if (save_instr[i] < 0) continue;
	  /*
	   * create a restore instruction by duplicating the appropriate
	   * save instruction and then changing the type of the duplicate
	   * from a store to a load.
	   */
	  iins = save_instr[i];
	  xins = AinsDup(iins);
	  rA = AINS_REG_A(iins);
	  rC = AINS_REG_A(iins);
	  AINS_CODE(xins) = REG_IS_INT(rA) ? I_ldq : I_ldt;
	  AINS_REG_A(xins) = rC;
	  AINS_REG_C(xins) = rA;
	  AINS_FLAGS(xins) &= ~(FI_EVAL_A|FI_EVAL_B|FI_EVAL_C|FI_DEF_A|FI_DEF_B);
	  AINS_BBL(xins) = jbbl;
	  /*
	   * skip past any sort of jump instruction at the end.
	   */
	  iins = ABBL_INS_LAST(jbbl);
	  if (VALID(iins)) {
	    if (AINS_TYPE(iins) == IT_IBR
		|| AINS_TYPE(iins) == IT_FBR
		|| AINS_TYPE(iins) == IT_BRA
		|| AINS_TYPE(iins) == IT_JMP) {
	      AinsInsertBefore(xins, iins);
	    }
	    else {
	      AinsInsertAfter(xins, iins);
	    }
	  }
	  else {
	    AinsAppend(xins);
	  }
	  winstrs -= ABBL_COUNT(jbbl);
	  delta   -= ABBL_COUNT(jbbl);
	  if (verbose) {
	    printf("\t\t[ShrinkWrap: +restore $%d: bbl %d, fun %d (ct = %ld)]\n", 
		   r, jbbl, ifun, ABBL_COUNT(jbbl));
	  }
	  chg = TRUE;
	}
      }
    }
  }

  /*
   * finally, kill unnecessary save and restore instructions.
   */
  FOREACH_FUN_BBL(ifun, ibbl) {
    if (ABBL_TYPE(ibbl) != BT_RET) continue;
    sregs = REGSET_DIFF(AbblRestoresRegs(ibbl), SmapToRegSet(ABBL_RESTORE(ibbl)));
    sregs = SmapToRegSet(RegSetToSmap(sregs));  /* mask out not-handled regs */
    if (sregs) {
      REGSET_FOREACH_REG(sregs, r) {
	i = r-8;
	if (save_instr[i] < 0) continue;

	FOREACH_BBL_INS_R(ibbl, iins) {
	  if ( (AINS_CODE(iins) == I_ldq || AINS_CODE(iins) == I_stt)
	       && AINS_REG_C(iins) == r) {
	    CASSERT((AINS_REG_B(iins) == REG_SP),
		    ("restore not from stack: reg %d [fun=%d]\n", r, ifun));
	    AinsKill(iins);
	    winstrs += ABBL_COUNT(ibbl);
	    delta   += ABBL_COUNT(ibbl);
	    if (verbose) {
	      printf("\t\t[ShrinkWrap: -restore $%d: bbl %d, fun %d (ct = %ld)]\n", 
		     r, ibbl, ifun, ABBL_COUNT(ibbl));
	    }
	    break;
	  }
	}
      }
    }
  }
  ibbl = AFUN_BBL_FIRST(ifun);
  sregs = REGSET_DIFF(AbblSavesRegs(ibbl), SmapToRegSet(ABBL_SAVE(ibbl)));
  sregs = SmapToRegSet(RegSetToSmap(sregs));  /* mask out not-handled regs */
  if (sregs) {
    REGSET_FOREACH_REG(sregs, r) {
      i = r-8;
      if (save_instr[i] < 0) continue;

      AinsKill(save_instr[i]);
      ninstrs += 1;
      winstrs += ABBL_COUNT(ibbl);
      delta   += ABBL_COUNT(ibbl);
      if (verbose) {
	printf("\t\t[ShrinkWrap: -save $%d: bbl %d, fun %d (ct = %ld)]\n", r, ibbl, ifun, ABBL_COUNT(ibbl));
      }
    }
  }
  CWARNING((delta < 0), ("Shrink-wrapping gives negative improvement: fun %d [wt chg = %ld]\n", ifun, delta));
}

/*=======================================================
@FUN
Shrink-wrap callee-saved registers where this is deemed safe.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptShrinkWrap"

extern BOOL OptShrinkWrap(FILE *fp)
{
  INDEX ifun;
  UINT8 sregmap;
  BOOL change = 0;
  clock_t time = clock();

  FOREACH_LIVE_FUN(ifun)
  {
    if( AFUN_IS_PSEUDO(ifun) ) continue;
    if( AFUN_HAS_STK_INVALID(ifun) ) continue;
    if( AfunIsSetjmp(ifun) ) continue;
    if( AfunIsLongjmp(ifun) ) continue;
    if( AfunHasEsc2(ifun) ) continue;

    if ( AfunInsertRetBbls(ifun) )
    {
      sregmap = ComShrinkWrapInfo(ifun);
      if (AfunShrinkWrapIsSafe(ifun, sregmap))
      {
	AfunShrinkWrap(ifun);
      }
    }
  }

  change = ninstrs > 0;
  time = (clock() - time) * 1000 / CLOCKS_PER_SEC;

  InfoStatus(fp, "\t[OptShrinkWrap%c] Moved Saves/Restores %d [weight chg %ld] {%d ms}\n",
             CHANGE(change), ninstrs, winstrs, (int)time);

   return change;
}


